vbBack = "\x08"
vbCr = "\x0d"
vbCrLf = "\x0d\x0a"
vbFormFeed = "\x0c"
vbLf = "\x0a"
vbNewLine = "\n"
vbNullChar = "\x00"
vbTab = "\x09"
vbVerticalTab = "\x0b"
vbNullString = ""
vbObjectError = -2147221504
